#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "config.h"

extern char *mysprintf(char *format, ...);
extern int errno;

int verbose               = 0;
int debug                 = 0;
int code_only             = 0;
int number_of_loft_files  = 0;

char **loft_files         = NULL;

static char *command      = "";
static char *libs         = "";
static char *loft_c_files = "";
static char *c_compiler   = "";
static int   executable   = 1;

static char *bin_dir, *lib_dir, *incl_dir;

/*****************************************************************/
#define FT       1
#define DIRECT   2
#define REWRITE  3
#define REPLACE  4
#define OPTIM    5
#define NETWORK  6
#define CONTROL  7
#define TSC      8

// default compiler translates into FairThreads
static int implem = FT; 

/*****************************************************************/
/* Processing the -version option */
static void print_version (void){
   printf ("--- loft: version %s (%s)\n",LOFT_VERSION,DATE);
}

/* is a .loft or .LOFT file ? */
static int is_loft_file(char *file){
   char *tail = file + strlen(file)-5;
   return (0 == strcmp (tail,".loft"))||(0 == strcmp (tail,".LOFT"));
}

/* option analysis */
#define ADD(s1,s2) s1 = mysprintf("%s %s ",s1,s2)

static void dispatch_options(int argc, char **argv){
   int i;
   for (i = 1 ; i < argc; i++) {
      if (*argv[i] == '-'){
         if (0 == strcmp(argv[i],"-version")){
            print_version ();
            exit(0);
         }

	 if (argv[i][0]=='-' &&
	     (argv[i][1]=='I' || argv[i][1]=='D' || argv[i][1]=='U'))
	 {
            incl_dir = mysprintf ("%s %s",incl_dir,argv[i]);
	    if (2 == strlen (argv[i])) {
	       incl_dir = mysprintf ("%s %s",incl_dir,argv[++i]);
	    }
         }else if (0 == strcmp(argv[i],"-v")){
            verbose = 1;
	 }else if (argv[i][0]=='-' && argv[i][1]=='l'){
            ADD(libs,argv[i]);
         }else if (0 == strcmp(argv[i],"-debug")){
            verbose = 1;
            debug = 1;
         }else if (0 == strcmp(argv[i],"-code")){
            code_only = 1;
         }else if (0 == strcmp(argv[i],"-o")){
            ADD(command,argv[i]);
            ADD(command,argv[++i]);
         }else if (0 == strcmp(argv[i],"-c")){
            ADD(command,argv[i]);
            executable = 0;
         }else if (0 == strcmp(argv[i],"-ft")){
            implem = FT;
         }else if (0 == strcmp(argv[i],"-direct")){
	    implem = DIRECT;
         }else if (0 == strcmp(argv[i],"-rewrite")){
	    implem = REWRITE;
         }else if (0 == strcmp(argv[i],"-replace")){
	    implem = REPLACE;
         }else if (0 == strcmp(argv[i],"-optim")){
	    implem = OPTIM;
	 }else if (0 == strcmp(argv[i],"-tsc")){
	    implem = TSC;
         }else if (0 == strcmp(argv[i],"-network")){
	    implem = NETWORK;
	 }else if (0 == strcmp(argv[i],"-control")){
	    implem = CONTROL;	    
         }else{	    
            ADD(command,argv[i]);
         }
      }else if (is_loft_file (argv[i])){
         loft_files[number_of_loft_files++] = argv[i];
      }else{
         ADD(command,argv[i]);
      }
   }
}

/* execution of a command using a system call */
static void execute (char *command){
   int res;
   if (verbose) fprintf (stderr,"%s\n",command);
   if (debug) return;
   if ((res = system (command))){
      fprintf (stderr,"cannot execute %s: %s\n",
	       command,strerror (errno));
      exit (res);
   }
}

/* base name of a .loft file */
char *base_loft_file (char *file)
{
   char *tmp,*res;
   res = mysprintf ("%s",file);
   tmp = strrchr (res,'/');
   return tmp ? tmp+1 : res;
}

/* cannot open file */
void open_error (char *name)
{
   fprintf (stderr,"*** loft: cannot open %s: %s\n",
	    name,strerror (errno));
   exit(1);
}

/************ main function ****************************/
int main (int argc, char **argv){
   int i;
   char *trans_command;
   char *libraries;
   char *directories;
   
   loft_files = malloc (argc * sizeof(char*));

   c_compiler = getenv ("C_COMPILER");
   if (!c_compiler) c_compiler = C_COMPILER;
   bin_dir   = getenv ("LOFT_BINDIR");
   if (!bin_dir) bin_dir = BIN_DIR;
   lib_dir   = getenv ("LOFT_LIBDIR");
   if (!lib_dir) lib_dir = LIB_DIR;
   incl_dir  = getenv ("LOFT_INCLDIR");
   if (!incl_dir) incl_dir = INCL_DIR;
   
   dispatch_options (argc,argv);

   if (implem == FT) {
      trans_command = "%s/loft2ft %s -I%s > %s";
      directories = mysprintf ("-I%s -L%s",incl_dir,lib_dir);
      libraries = "-lloft_ft -lfthread -lpthread";
   } else {
      directories = mysprintf ("-I%s -L%s",incl_dir,lib_dir);      
      if (implem == DIRECT) {
	 trans_command = "%s/loft2c %s -I%s > %s";
	 libraries = "-lloft_c";
      } else if (implem == REWRITE) {
	 trans_command = "%s/loft2sem %s -I%s > %s";
	 libraries = "-lloft_rewrite";
      } else if (implem == REPLACE) {
	 trans_command = "%s/loft2sem %s -I%s > %s";
	 libraries = "-lloft_replace";
      } else if (implem == OPTIM) {
	 trans_command = "%s/loft2sem %s -I%s > %s";
	 libraries = "-lloft_optim";
      } else if (implem == TSC) {
	 trans_command = "%s/loft2tsc %s -I%s > %s";
	 libraries = "-lloft_tsc -lpthread";
      } else if (implem == NETWORK) {
	 trans_command = "%s/loft2tsc %s -I%s > %s";
	 libraries = "-lloft_net -ldl";
      } else if (implem == CONTROL) {
	 trans_command = "%s/loft2tsc %s -I%s > %s";
	 libraries = "-lloft_ctrl -ldl -lpthread";
      }        
   }
   
   for(i = 0; i < number_of_loft_files; i++) {
      char *file = loft_files[i];
      if (0==freopen (file,"r",stdin)) open_error(file);
      else {
	 char *base_c = mysprintf ("%s.c",base_loft_file (file));
         execute (mysprintf (trans_command,bin_dir,file,incl_dir,base_c));
         ADD (loft_c_files,mysprintf ("%s.c",base_loft_file (file)));
      }
   }
   
   command = mysprintf ("%s %s %s %s %s %s",
     c_compiler,
     directories,
     command,
     loft_c_files,
     libs,
     executable ? libraries : ""
   );
   
   if (!code_only) execute(command);

   exit (0);
   return 0;
}
