#include "rewrite_sem.h"

rewrite_result_t make_rewrite_result (int status, instruction_t inst)
{
   rewrite_result_t res = mymalloc (sizeof (struct rewrite_result_t));
   res->status = status;
   res->inst = inst;
   return res;
}

/******************************************************/
typedef rewrite_result_t (*rewrite_fun_t)(instruction_t);
rewrite_fun_t dispach_rewrite [] =
{
      nothing_rewrite             ,
      cooperate_rewrite           ,
      halt_rewrite                ,
      return_rewrite              ,
      sequence_rewrite            ,
      stop_rewrite                ,
      suspend_rewrite             ,
      resume_rewrite              ,
      if_rewrite                  ,
      expression_rewrite          ,
      repeat_rewrite              ,
      generate_rewrite            ,
      generate_valued_rewrite     ,
      await_rewrite               ,
      limited_await_rewrite       ,
      atom_rewrite                ,
      join_rewrite                ,
      limited_join_rewrite        ,
      run_rewrite                 ,
      while_rewrite               ,
      get_value_rewrite           ,
      repeat_star_rewrite         ,
      while_star_rewrite          ,
      limited_await_star_rewrite  ,
      await_star_rewrite          ,
      get_value_star_rewrite      ,
};

rewrite_result_t rewrite (instruction_t inst)
{
   return dispach_rewrite [(int)inst->type] (inst);
}

/****************************************************/
void fire_thread (thread_t thread)
{
   rewrite_result_t res;
   thread->scheduler->current = thread; // set self
   res = rewrite (thread->run);
   return_processing (thread,res->status);
   thread->run = res->inst;
}

/******************************************************/
rewrite_result_t nothing_rewrite (instruction_t inst)
{
   return make_rewrite_result (_TERM,make_nothing ());
}

rewrite_result_t cooperate_rewrite (instruction_t inst)
{
   return make_rewrite_result (_COOP,make_nothing ());
}

rewrite_result_t halt_rewrite (instruction_t inst)
{
   return make_rewrite_result (_COOP,make_halt ());
}

rewrite_result_t return_rewrite (instruction_t inst)
{
   broadcast (self ()->signal);
   return make_rewrite_result (_RTRN,make_nothing ());
}

/***************************************************/
rewrite_result_t sequence_rewrite (instruction_t inst)
{
   sequence_t s = (sequence_t)inst;
   rewrite_result_t res = rewrite (s->left);
   if (res->status != _TERM) {
      return make_rewrite_result (
	       res->status,
	       make_sequence(res->inst,s->right));
   }
   return rewrite(s->right);
}

/***************************************************/
rewrite_result_t stop_rewrite (instruction_t inst)
{
   stop_t s = (stop_t)inst;
   thread_t th = evaluate (s->exp);
   stop (th);
   return make_rewrite_result (_TERM,make_nothing ());   
}

/***************************************************/
rewrite_result_t suspend_rewrite (instruction_t inst)
{
   suspend_t s = (suspend_t)inst;
   thread_t th = evaluate (s->exp);
   suspend (th);
   return make_rewrite_result (_TERM,make_nothing ());   
}

/***************************************************/
rewrite_result_t resume_rewrite (instruction_t inst)
{
   resume_t r = (resume_t)inst;
   thread_t th = evaluate (r->exp);   
   resume (th);
   return make_rewrite_result (_TERM,make_nothing ());   
}

/***************************************************/
rewrite_result_t if_rewrite (instruction_t inst)
{
   if_t test = (if_t)inst;
   int b = (int)evaluate (test->exp);
   if (b) return rewrite (test->left);
   else return rewrite (test->right);   
}

/***************************************************/
rewrite_result_t expression_rewrite (instruction_t inst)
{
   inst_from_exp_t i = (inst_from_exp_t)inst;
   evaluate (i->exp);
   return make_rewrite_result (_TERM,make_nothing ());
}

/***************************************************/
rewrite_result_t repeat_rewrite (instruction_t inst)
{
   repeat_t r = (repeat_t)inst;
   int n = (int)evaluate (r->exp);
   return rewrite (make_repeat_star (n,r->body,r->body));
}

/***************************************************/
rewrite_result_t repeat_star_rewrite (instruction_t inst)
{
   repeat_star_t r = (repeat_star_t)inst;   
   rewrite_result_t res;

   if (r->counter <= 0)
      return make_rewrite_result (_TERM,make_nothing ());
   
   res = rewrite (r->body);
   
   if (res->status != _TERM)
      return make_rewrite_result (
	        res->status,
	        make_repeat_star (r->counter,r->init,res->inst));

   return rewrite (make_repeat_star (r->counter-1,r->init,r->init));   
}

/***************************************************/
rewrite_result_t generate_rewrite (instruction_t inst)
{
   generate_t gen = (generate_t)inst;
   event_t event = evaluate (gen->exp);
   generate (event);
   return make_rewrite_result (_TERM,make_nothing ());
}


/***************************************************/
rewrite_result_t generate_valued_rewrite (instruction_t inst)
{
   generate_valued_t gen = (generate_valued_t)inst;
   event_t event = evaluate (gen->evt_exp);
   int *val = evaluate (gen->val_exp);
   generate_value (event,val);
   return make_rewrite_result (_TERM,make_nothing ());
}

/***************************************************/
rewrite_result_t await_rewrite (instruction_t inst)
{
   await_t a = (await_t)inst;
   event_t event = evaluate (a->exp);
   // the event scheduler must be the current scheduler
   return rewrite (make_await_star (event));
}

rewrite_result_t await_star_rewrite (instruction_t inst)
{
   await_star_t a = (await_star_t)inst;
   if (is_present (a->event))
      return make_rewrite_result (_TERM,make_nothing ());
   return make_rewrite_result (
      (current_scheduler ()->eoi ? _COOP : _CONT),
      make_await_star (a->event));
}

/***************************************************/
rewrite_result_t limited_await_rewrite (instruction_t inst)
{
   limited_await_t a = (limited_await_t)inst;
   event_t event = evaluate (a->evt_exp);
   // the event scheduler must be the current scheduler   
   int n = (int)evaluate (a->limit_exp);
   
   return rewrite (make_limited_await_star (event,n));
}

rewrite_result_t limited_await_star_rewrite (instruction_t inst)
{
   limited_await_star_t a = (limited_await_star_t)inst;
   
   if (a->limit <= 0) {
      self ()->code = ETIMEOUT;
      return make_rewrite_result (_TERM,make_nothing ());
   }
   if (is_present (a->event)) {
      self ()->code = OK;
      return make_rewrite_result (_TERM,make_nothing ());
   }
   if (current_scheduler()->eoi) {
      return make_rewrite_result (
	      _COOP,
	      make_limited_await_star (a->event,a->limit-1));
   }
   return make_rewrite_result (
              _CONT,
	      make_limited_await_star (a->event,a->limit));
}

/***************************************************/
rewrite_result_t atom_rewrite (instruction_t inst)
{
   atom_t a = (atom_t)inst;
   a->selector (a->n,self ()->locals);
   return make_rewrite_result (_TERM,make_nothing ());
}

/***************************************************/
rewrite_result_t join_rewrite (instruction_t inst)
{
   join_t j = (join_t)inst;
   thread_t thread = evaluate (j->exp);
   // the thread scheduler must be the current scheduler   
   if (thread->state == _TERM)
      return make_rewrite_result (_TERM,make_nothing ());
   return rewrite (make_await_star (thread->signal));
}

/***************************************************/
rewrite_result_t limited_join_rewrite (instruction_t inst)
{
   limited_join_t j = (limited_join_t)inst;
   thread_t thread = evaluate (j->th_exp);
   int n = (int)evaluate (j->limit_exp);
   // the thread scheduler must be the current scheduler   
   if (thread->state == _TERM)
      return make_rewrite_result (_TERM,make_nothing ());
   return rewrite (make_limited_await_star (thread->signal,n));
}

/***************************************************/
rewrite_result_t run_rewrite (instruction_t inst)
{
   run_t r = (run_t)inst;
   thread_t thread = evaluate (r->exp);
   self ()->called = thread;
   return rewrite (make_join (make_const_expression (thread)));
}

/***************************************************/
rewrite_result_t while_rewrite (instruction_t inst)
{
   while_t w = (while_t)inst;
   int b = (int)evaluate (w->exp);
   if (!b) return make_rewrite_result (_TERM,make_nothing ());
   return rewrite (make_while_star (w->exp,w->body,w->body));
}

rewrite_result_t while_star_rewrite (instruction_t inst)
{
   while_star_t w = (while_star_t)inst;

   rewrite_result_t res = rewrite (w->body);

   if (res->status != _TERM)
      return make_rewrite_result (
	 res->status,
	 make_while_star (w->exp,w->init,res->inst));

   return rewrite (make_while (w->exp,w->init));   
}

/***************************************************/
rewrite_result_t get_value_rewrite (instruction_t inst)
{
   get_value_t g = (get_value_t)inst;
   event_t event =  evaluate (g->evt_exp);
   int k = (int)evaluate (g->num_exp);
   // the event scheduler must be the current scheduler      
   return rewrite (make_get_value_star (event,k,g->var));   
}

rewrite_result_t get_value_star_rewrite (instruction_t inst)
{
   get_value_star_t g = (get_value_star_t)inst;
   int len = event_values_length (g->event);
   
   if (len > g->num) {
      (*g->var) = nth_of_event_values (g->event,g->num);
      self ()->code = OK;      
      return make_rewrite_result (_TERM,make_nothing ());
   }
   if (current_scheduler ()->eoi) {
      self ()->code = ENEXT;
      return make_rewrite_result (_COOP,make_nothing ());
   }
   
   return make_rewrite_result (_CONT,
	       make_get_value_star (g->event,g->num,g->var));
}

